from .jwt_config import (
    JwtConfig, JwtDbConfig, JwtAlgorithm
)
from .jwt_pomes import (
    jwt_needed, jwt_verify_request,
    jwt_assert_account, jwt_set_account, jwt_remove_account,
    jwt_issue_token, jwt_issue_tokens, jwt_refresh_tokens,
    jwt_get_claims, jwt_validate_token, jwt_revoke_token
)
from .jwt_providers import (
    provider_register, provider_get_token
)

__all__ = [
    # jwt_constants
    "JwtConfig", "JwtDbConfig", "JwtAlgorithm",
    # jwt_pomes
    "jwt_needed", "jwt_verify_request",
    "jwt_assert_account", "jwt_set_account", "jwt_remove_account",
    "jwt_issue_token", "jwt_issue_tokens", "jwt_refresh_tokens",
    "jwt_get_claims", "jwt_validate_token", "jwt_revoke_token",
    # jwt_providers
    "provider_register", "provider_get_token"
]

from importlib.metadata import version
__version__ = version("pypomes_jwt")
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())



from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from enum import Enum, StrEnum
from pypomes_core import (
    APP_PREFIX,
    env_get_str, env_get_bytes, env_get_int, env_get_enum
)
from secrets import token_bytes


class JwtAlgorithm(StrEnum):
    """
    Supported decoding algorithms.
    """
    HS256 = "HS256"
    HS512 = "HS512"
    RS256 = "RS256"
    RS512 = "RS512"


# recommended: allow the encode and decode keys to be generated anew when app starts
_encoding_key: bytes = env_get_bytes(key=f"{APP_PREFIX}_JWT_ENCODING_KEY",
                                     encoding="base64url")
_decoding_key: bytes
_default_algorithm: JwtAlgorithm = env_get_enum(key=f"{APP_PREFIX}_JWT_DEFAULT_ALGORITHM",
                                                enum_class=JwtAlgorithm,
                                                def_value=JwtAlgorithm.RS256)
if _default_algorithm in [JwtAlgorithm.HS256, JwtAlgorithm.HS512]:
    if not _encoding_key:
        _encoding_key = token_bytes(nbytes=32)
    _decoding_key = _encoding_key
else:
    _decoding_key: bytes = env_get_bytes(key=f"{APP_PREFIX}_JWT_DECODING_KEY")
    if not _encoding_key or not _decoding_key:
        __priv_key: RSAPrivateKey = rsa.generate_private_key(public_exponent=65537,
                                                             key_size=2048)
        _encoding_key = __priv_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                 format=serialization.PrivateFormat.PKCS8,
                                                 encryption_algorithm=serialization.NoEncryption())
        __pub_key: RSAPublicKey = __priv_key.public_key()
        _decoding_key = __pub_key.public_bytes(encoding=serialization.Encoding.PEM,
                                               format=serialization.PublicFormat.SubjectPublicKeyInfo)


class JwtConfig(Enum):
    """
    Parameters for JWT token issuance.
    """
    # recommended: between 5 min and 1 hour (set to 5 min)
    ACCESS_MAX_AGE: int = env_get_int(key=f"{APP_PREFIX}_JWT_ACCESS_MAX_AGE",
                                      def_value=300)
    ACCOUNT_LIMIT: int = env_get_int(key=f"{APP_PREFIX}_JWT_ACCOUNT_LIMIT",
                                     def_value=5)
    DEFAULT_ALGORITHM: _default_algorithm
    ENCODING_KEY: bytes = _encoding_key
    DECODING_KEY: bytes = _decoding_key
    # recommended: at least 2 hours (set to 24 hours)
    REFRESH_MAX_AGE: int = env_get_int(key=f"{APP_PREFIX}_JWT_REFRESH_MAX_AGE",
                                       def_value=86400)


del _decoding_key
del _encoding_key
del _default_algorithm


class JwtDbConfig(StrEnum):
    """
    Parameters for JWT database connection.
    """
    ENGINE = env_get_str(key=f"{APP_PREFIX}_JWT_DB_ENGINE")
    TABLE = env_get_str(key=f"{APP_PREFIX}_JWT_DB_TABLE")
    COL_ACCOUNT = env_get_str(key=f"{APP_PREFIX}_JWT_DB_COL_ACCOUNT")
    COL_ALGORITHM = env_get_str(key=f"{APP_PREFIX}_JWT_DB_COL_ALGORITHM")
    COL_DECODER = env_get_str(key=f"{APP_PREFIX}_JWT_DB_COL_DECODER")
    COL_KID = env_get_str(key=f"{APP_PREFIX}_JWT_DB_COL_KID")
    COL_TOKEN = env_get_str(key=f"{APP_PREFIX}_JWT_DB_COL_TOKEN")



import jwt
import string
import sys
from base64 import b64encode
from datetime import datetime, UTC
from logging import Logger
from pypomes_core import str_random
from pypomes_db import (
    DbEngine, db_connect, db_commit, db_rollback, db_close,
    db_select, db_insert, db_update, db_delete
)
from threading import Lock
from typing import Any

from .jwt_config import JwtConfig, JwtDbConfig


class JwtRegistry:
    """
    Shared JWT registry for security token access.

    Instance variables:
      - access_lock: lock for safe multi-threading access
      - access_data: dictionary holding the JWT token data, organized by account id:
       {
         <account-id>: {
           "access-max-age": <int>,         # defaults to JWT_ACCESS_MAX_AGE (in seconds)
           "refresh-max-age": <int>,        # defaults to JWT_REFRESH_MAX_AGE (in seconds)
           "lead-interval": <int>,          # time to wait for token to be valid, in seconds
           "claims": {
             "iss": <string>,               # token'ss issuer
             "birthdate": <string>,         # subject's birth date
             "email": <string>,             # subject's email
             "gender": <string>,            # subject's gender
             "name": <string>,              # subject's name
             "roles": <List[str]>,          # subject roles
             "nonce": <string>,             # used to associate a Client session with a token
             ...
           }
         },
         ...
       }

    JSON Web Token (JWT) is a compact, URL-safe means of representing claims to be transferred between
    two parties. It is fully described in the RFC 7519, issued by the Internet Engineering Task Force
    (see https://www.rfc-editor.org/rfc/rfc7519.html).
    In this context, claims are pieces of information a token bears, and herein are loosely classified
    as token-related and account-related. All times are UTC.

    Token-related claims are mostly required claims, and convey information about the token itself:
      # required
      "exp": <timestamp>        expiration time
      "iat": <timestamp>        issued at
      "iss": <string>           token's issuer
      "jti": <string>           JWT id
      "sub": <string>           subject (the account identification)
      # optional
      "aud": <string>           token audience
      "nbt": <timestamp>        not before time

    Account-related claims are optional claims, and convey information about the registered account they belong to.
    Alhough they can be freely specified, these are some of the most commonly used claims:
       "birthdate": <string>    subject's birth date
       "email": <string>        subject's email
       "gender": <string>       subject's gender
       "name": <string>         subject's name
       "roles": <List[str]>     subject roles
       "nonce": <string>        used to associate a client session with a token

    The token header has these items:
      "alg": <string>           the algorithm used to sign the token (one of *HS256*, *HS51*', *RSA256*, *RSA512*)
      "typ": <string>           the token type (fixed to *JWT*)
      "kid": <string>           a token type and key to its location in the token database

    If issued by the local server, "kid" holds the key to the corresponding record in the token database,
    if starting with *A* for (*Access*) or *R* (for *Refresh*), followed an integer.
    """
    def __init__(self) -> None:
        """
        Initizalize the token access data.
        """
        self.access_lock: Lock = Lock()
        self.access_registry: dict[str, Any] = {}

    def add_account(self,
                    account_id: str,
                    claims: dict[str, Any],
                    access_max_age: int,
                    refresh_max_age: int,
                    lead_interval: int | None,
                    logger: Logger = None) -> None:
        """
        Add to storage the parameters needed to produce and validate JWT tokens for *account_id*.

        The parameter *claims* may contain account-related claims, only. Ideally, it should contain,
        at a minimum, *iss*, *birthdate*, *email*, *gender*, *name*, and *roles*.
        The parameter *refresh_max_age* should be at least 300 seconds greater than *access-max-age*.

        :param account_id: the account identification
        :param claims: the JWT claimset, as key-value pairs
        :param access_max_age: access token duration, in seconds (at least 60 seconds)
        :param refresh_max_age: refresh token duration, in seconds (greater than *access_max_age*)
        :param lead_interval: time to wait for token to be valid, in seconds
        :param logger: optional logger
        """
        # build and store the access data for the account
        with self.access_lock:
            if account_id not in self.access_registry:
                self.access_registry[account_id] = {
                    "access-max-age": access_max_age,
                    "refresh-max-age": refresh_max_age,
                    "lead-interval": lead_interval,
                    "claims": claims or {}
                }
                if logger:
                    logger.debug(f"JWT data added for '{account_id}'")
            elif logger:
                logger.warning(f"JWT data already exists for '{account_id}'")

    def remove_account(self,
                       account_id: str,
                       logger: Logger) -> bool:
        """
        Remove from storage the access data for *account_id*.

        :param account_id: the account identification
        :param logger: optional logger
        return: *True* if the access data was removed, *False* otherwise
        """
        # remove from internal storage
        account_data: dict[str, Any] | None
        with self.access_lock:
            account_data = self.access_registry.pop(account_id, None)

        # remove from database
        db_delete(delete_stmt=f"DELETE FROM {JwtDbConfig}",
                  where_data={JwtDbConfig.COL_ACCOUNT: account_id},
                  engine=DbEngine(JwtDbConfig.ENGINE),
                  logger=logger)
        if logger:
            if account_data:
                logger.debug(f"Removed JWT data for '{account_id}'")
            else:
                logger.warning(f"No JWT data found for '{account_id}'")

        return account_data is not None

    def issue_token(self,
                    account_id: str,
                    nature: str,
                    duration: int,
                    lead_interval: int = None,
                    claims: dict[str, Any] = None,
                    logger: Logger = None) -> str:
        """
        Issue an return a JWT token associated with *account_id*.

        The parameter *nature* must be a single letter in the range *[B-Z]*, less *R*
        (*A* is reserved for *access* tokens, and *R* for *refresh* tokens).
        The parameter *duration* specifies the token's validity interval (at least 60 seconds).
        These claims are ignored, if specified in *claims*: *iat*, *iss*, *exp*, *jti*, *nbf*, and *sub*.

        :param account_id: the account identification
        :param nature: the token's nature, must be a single letter in the range *[B-Z]*, less *R*
        :param duration: the number of seconds for the token to remain valid (at least 60 seconds)
        :param claims: optional token's claims
        :param lead_interval: optional interval for the token to become active (in seconds)
        :param logger: optional logger
        :return: the JWT token
        :raises RuntimeError: invalid parameter
        """
        # validate some parameters
        err_msg: str | None = None
        if not isinstance(nature, str) or \
                len(nature) != 1 or nature < "A" or nature > "Z":
            err_msg: str = f"Invalid nature '{nature}'"
        elif not isinstance(duration, int) or duration < 60:
            err_msg = f"Invalid duration '{duration}'"
        if err_msg:
            if logger:
                logger.error(err_msg)
            raise RuntimeError(err_msg)

        # obtain the account data in storage (may raise an exception)
        account_data: dict[str, Any] = self.get_account_data(account_id=account_id,
                                                             logger=logger)
        # issue the token
        current_claims: dict[str, Any] = {}
        iss: str = account_data["claims"].get("iss")
        if iss:
            current_claims["iss"] = iss
        if claims:
            current_claims.update(claims)

        current_claims["jti"] = str_random(size=32,
                                           chars=string.ascii_letters + string.digits)
        current_claims["sub"] = account_id
        just_now: int = int(datetime.now(tz=UTC).timestamp())
        current_claims["iat"] = just_now
        if lead_interval:
            current_claims["nbf"] = just_now + lead_interval
        current_claims["exp"] = just_now + duration

        # may raise an exception
        return jwt.encode(payload=current_claims,
                          key=JwtConfig.ENCODING_KEY.value,
                          algorithm=JwtConfig.DEFAULT_ALGORITHM.value,
                          headers={"kid": nature})

    def issue_tokens(self,
                     account_id: str,
                     account_claims: dict[str, Any] = None,
                     db_conn: Any = None,
                     logger: Logger = None) -> dict[str, Any]:
        """
        Issue and return a JWT token pair associated with *account_id*.

        These claims are ignored, if specified in *account_claims*: *iat*, *exp*, *jti*, *nbf*, and *sub*.
        Other claims specified therein may supercede registered account-related claims.

        If provided, *db_conn* indicates that this operation is part of a larger database transaction.
        Otherwise, the database transaction's scope is limited to this operation.

        Structure of the return data:
        {
          "access-token": <jwt-token>,
          "created-in": <timestamp>,
          "expires-in": <seconds-to-expiration>,
          "refresh-token": <jwt-token>
        }

        :param account_id: the account identification
        :param account_claims: if provided, may supercede registered account-related claims
        :param db_conn: if provided, indicates that this operation is part of a larger database transaction
        :param logger: optional logger
        :return: the JWT token data
        :raises RuntimeError: invalid account id, or error accessing the token database
        """
        # process the account data in storage
        with (self.access_lock):
            account_data: dict[str, Any] = self.get_account_data(account_id=account_id,
                                                                 logger=logger)
            current_claims: dict[str, Any] = account_data["claims"].copy()
            if account_claims:
                current_claims.update(account_claims)
            current_claims["jti"] = str_random(size=32,
                                               chars=string.ascii_letters + string.digits)
            current_claims["sub"] = account_id
            errors: list[str] = []

            just_now: int = int(datetime.now(tz=UTC).timestamp())
            current_claims["iat"] = just_now
            lead_interval = account_data.get("lead-interval")
            if lead_interval:
                current_claims["nbf"] = just_now + lead_interval

            # issue a candidate refresh token first, and persist it
            current_claims["exp"] = just_now + account_data.get("refresh-max-age")
            # may raise an exception
            refresh_token: str = jwt.encode(payload=current_claims,
                                            key=JwtConfig.ENCODING_KEY.value,
                                            algorithm=JwtConfig.DEFAULT_ALGORITHM.value,
                                            headers={"kid": "R0"})

            # make sure to have a database connection
            curr_conn: Any = db_conn or db_connect(autocommit=False,
                                                   engine=DbEngine(JwtDbConfig.ENGINE),
                                                   errors=errors,
                                                   logger=logger)
            if curr_conn:
                # persist the candidate token (may raise an exception)
                token_id: int = JwtRegistry.jwt_persist_token(account_id=account_id,
                                                              jwt_token=refresh_token,
                                                              db_conn=curr_conn,
                                                              logger=logger)
                # issue the definitive refresh token
                refresh_token = jwt.encode(payload=current_claims,
                                           key=JwtConfig.ENCODING_KEY.value,
                                           algorithm=JwtConfig.DEFAULT_ALGORITHM.value,
                                           headers={"kid": f"R{token_id}"})
                # persist it
                db_update(update_stmt=f"UPDATE {JwtDbConfig.TABLE}",
                          update_data={JwtDbConfig.COL_TOKEN: refresh_token},
                          where_data={JwtDbConfig.COL_KID: token_id},
                          engine=DbEngine(JwtDbConfig.ENGINE),
                          connection=curr_conn,
                          errors=errors,
                          logger=logger)

                # wrap-up the transaction
                if not db_conn:
                    if errors:
                        db_rollback(connection=curr_conn,
                                    logger=logger)
                    else:
                        db_commit(connection=curr_conn,
                                  errors=errors,
                                  logger=logger)
                    db_close(connection=curr_conn)

            if errors:
                raise RuntimeError("; ".join(errors))

            # issue the access token
            current_claims["exp"] = just_now + account_data.get("access-max-age")
            # may raise an exception
            access_token: str = jwt.encode(payload=current_claims,
                                           key=JwtConfig.ENCODING_KEY.value,
                                           algorithm=JwtConfig.DEFAULT_ALGORITHM.value,
                                           headers={"kid": f"A{token_id}"})
            # return the token data
            return {
                "access-token": access_token,
                "created-in": current_claims.get("iat"),
                "expires-in": current_claims.get("exp"),
                "refresh-token": refresh_token
            }

    def get_account_data(self,
                         account_id: str,
                         logger: Logger = None) -> dict[str, Any]:
        """
        Retrieve the JWT access data associated with *account_id*.

        :return: the JWT access data associated with *account_id*
        :raises RuntimeError: No JWT access data exists for *account_id*
        """
        # retrieve the access data
        result: dict[str, Any] = self.access_registry.get(account_id)
        if not result:
            # JWT access data not found
            err_msg: str = f"No JWT access data found for '{account_id}'"
            if logger:
                logger.error(err_msg)
            raise RuntimeError(err_msg)

        return result

    @staticmethod
    def jwt_persist_token(account_id: str,
                          jwt_token: str,
                          db_conn: Any = None,
                          logger: Logger = None) -> int:
        """
        Persist the given token, making sure that the account limit is complied with.

        The tokens in storage, associated with *account_id*, are examined for their expiration timestamp.
        If a token's expiration timestamp is in the past, it is removed from storage. If the maximum number
        of active tokens for *account_id* has been reached, the oldest active one is alse removed,
        to make room for the new *jwt_token*.

        If provided, *db_conn* indicates that this operation is part of a larger database transaction.
        Otherwise, the database transaction's scope is limited to this operation.

        :param account_id: the account identification
        :param jwt_token: the JWT token to persist
        :param db_conn: the database connection to use
        :param logger: optional logger
        :return: the storage id of the inserted token
        :raises RuntimeError: error accessing the token database
        """
        from .jwt_pomes import jwt_get_claims

        # initialize the return variable
        result: int | None = None

        # make sure to have a database connection
        errors: list[str] = []
        curr_conn: Any = db_conn or db_connect(autocommit=False,
                                               engine=DbEngine(JwtDbConfig.ENGINE),
                                               errors=errors,
                                               logger=logger)
        if not errors:

            # retrieve the account's tokens
            # noinspection PyTypeChecker
            recs: list[tuple[int, str, str, str]] = \
                db_select(sel_stmt=f"SELECT {JwtDbConfig.COL_KID}, {JwtDbConfig.COL_TOKEN} "
                                   f"FROM {JwtDbConfig.TABLE}",
                          where_data={JwtDbConfig.COL_ACCOUNT: account_id},
                          engine=DbEngine(JwtDbConfig.ENGINE),
                          connection=curr_conn,
                          errors=errors,
                          logger=logger)
            if not errors:
                if logger:
                    logger.debug(msg=f"Retrieved {len(recs)} tokens from storage for account '{account_id}'")
                # remove the expired tokens
                just_now: int = int(datetime.now(tz=UTC).timestamp())
                oldest_ts: int = sys.maxsize
                oldest_id: int | None = None
                expired: list[int] = []
                for rec in recs:
                    token: str = rec[1]
                    token_id: int = rec[0]
                    token_payload: dict[str, Any] = (jwt_get_claims(token=token,
                                                                    errors=errors,
                                                                    logger=logger) or {}).get("payload")
                    if errors:
                        break
                    # find expired tokens
                    exp: int = token_payload.get("exp", sys.maxsize)
                    if exp < just_now:
                        expired.append(token_id)

                    # find oldest token
                    iat: int = token_payload.get("iat", sys.maxsize)
                    if iat < oldest_ts:
                        oldest_ts = iat
                        oldest_id = token_id

                # remove expired tokens from persistence
                if not errors and expired:
                    db_delete(delete_stmt=f"DELETE FROM {JwtDbConfig.TABLE}",
                              where_data={JwtDbConfig.COL_KID: expired},
                              engine=DbEngine(JwtDbConfig.ENGINE),
                              connection=curr_conn,
                              errors=errors,
                              logger=logger)
                    if not errors and logger:
                        logger.debug(msg=f"{len(expired)} tokens of account "
                                         f"'{account_id}' removed from storage")

                if not errors and 0 < JwtConfig.ACCOUNT_LIMIT.value <= len(recs) - len(expired):
                    # delete the oldest token to make way for the new one
                    db_delete(delete_stmt=f"DELETE FROM {JwtDbConfig.TABLE}",
                              where_data={JwtDbConfig.COL_KID: oldest_id},
                              engine=DbEngine(JwtDbConfig.ENGINE),
                              connection=curr_conn,
                              errors=errors,
                              logger=logger)
                    if not errors and logger:
                        logger.debug(msg="Oldest active token of account "
                                         f"'{account_id}' removed from storage")
                # persist token
                if not errors:
                    result = db_insert(insert_stmt=f"INSERT INTO {JwtDbConfig.TABLE}",
                                       insert_data={
                                           JwtDbConfig.COL_ACCOUNT: account_id,
                                           JwtDbConfig.COL_TOKEN: jwt_token,
                                           JwtDbConfig.COL_ALGORITHM: JwtConfig.DEFAULT_ALGORITHM.value,
                                           JwtDbConfig.COL_DECODER: b64encode(s=JwtConfig.DECODING_KEY.value).decode()
                                       },
                                       return_cols={JwtDbConfig.COL_KID: int},
                                       engine=DbEngine(JwtDbConfig.ENGINE),
                                       connection=curr_conn,
                                       errors=errors,
                                       logger=logger)
        # finish the operation
        if not db_conn:
            if errors:
                db_rollback(connection=curr_conn)
            else:
                db_commit(connection=curr_conn)
            db_close(connection=curr_conn)

        if errors:
            raise RuntimeError("; ".join(errors))

        return result



import jwt
import sys
from base64 import b64decode
from flask import Request, Response, request
from logging import Logger
from pypomes_core import exc_format
from pypomes_db import (
    DbEngine, db_connect, db_commit,
    db_rollback, db_close, db_select, db_delete
)
from typing import Any

from .jwt_config import JwtConfig, JwtDbConfig
from .jwt_registry import JwtRegistry

# the JWT registry
__jwt_registry: JwtRegistry = JwtRegistry()


def jwt_needed(func: callable) -> callable:
    """
    Create a decorator to authenticate service endpoints with JWT tokens.

    :param func: the function being decorated
    """
    # ruff: noqa: ANN003 - Missing type annotation for *{name}
    def wrapper(*args, **kwargs) -> Response:
        response: Response = jwt_verify_request(request=request)
        return response if response else func(*args, **kwargs)

    # prevent a rogue error ("View function mapping is overwriting an existing endpoint function")
    wrapper.__name__ = func.__name__

    return wrapper


def jwt_verify_request(request: Request) -> Response:
    """
    Verify whether the HTTP *request* has the proper authorization, as per the JWT standard..

    This implementation assumes that HTTP requests are handled with the *Flask* framework.

    :param request: the *request* to be verified
    :return: *None* if the *request* is valid, otherwise a *Response* reporting the error
    """
    # initialize the return variable
    result: Response | None = None

    # retrieve the authorization from the request header
    auth_header: str = request.headers.get("Authorization")

    # validate the authorization token
    bad_token: bool = True
    if auth_header and auth_header.startswith("Bearer "):
        # yes, extract and validate the JWT access token
        token: str = auth_header.split(" ")[1]
        claims: dict[str, Any] = jwt_validate_token(token=token,
                                                    nature="A")
        if claims:
            login: str = request.values.get("login")
            subject: str = claims["payload"].get("sub")
            if not login or not subject or login == subject:
                bad_token = False

    # deny the authorization
    if bad_token:
        result = Response(response="Authorization failed",
                          status=401)
    return result


def jwt_assert_account(account_id: str) -> bool:
    """
    Determine whether access for *account_id* has been established.

    :param account_id: the account identification
    :return: *True* if access data exists for *account_id*, *False* otherwise
    """
    return __jwt_registry.access_registry.get(account_id) is not None


def jwt_set_account(account_id: str,
                    claims: dict[str, Any],
                    access_max_age: int = JwtConfig.ACCESS_MAX_AGE.value,
                    refresh_max_age: int = JwtConfig.REFRESH_MAX_AGE.value,
                    lead_interval: int = None,
                    logger: Logger = None) -> None:
    """
    Establish the data needed to obtain JWT tokens for *account_id*.

    The parameter *claims* may contain account-related claims, only. Ideally, it should contain,
    at a minimum, *iss*, *birthdate*, *email*, *gender*, *name*, and *roles*.
    It is enforced that the parameter *refresh_max_age* should be at least 300 seconds greater
    than *access-max-age*.

    :param account_id: the account identification
    :param claims: the JWT claimset, as key-value pairs
    :param access_max_age: access token duration, in seconds
    :param refresh_max_age: refresh token duration, in seconds
    :param lead_interval: optional time to wait for token to be valid, in seconds
    :param logger: optional logger
    """
    if logger:
        logger.debug(msg=f"Registering account data for '{account_id}'")

    # register the JWT service
    __jwt_registry.add_account(account_id=account_id,
                               claims=claims,
                               access_max_age=access_max_age,
                               refresh_max_age=max(refresh_max_age, access_max_age + 300),
                               lead_interval=lead_interval,
                               logger=logger)


def jwt_remove_account(account_id: str,
                       logger: Logger = None) -> bool:
    """
    Remove from storage the JWT access data for *account_id*.

    :param account_id: the account identification
    :param logger: optional logger
    return: *True* if the access data was removed, *False* otherwise
    """
    if logger:
        logger.debug(msg=f"Remove access data for '{account_id}'")

    return __jwt_registry.remove_account(account_id=account_id,
                                         logger=logger)


def jwt_validate_token(token: str,
                       nature: str = None,
                       account_id: str = None,
                       errors: list[str] = None,
                       logger: Logger = None) -> dict[str, Any] | None:
    """
    Verify if *token* is a valid JWT token.

    Attempt to validate non locally issued tokens will not succeed. If *nature* is provided,
    validate whether *token* is of that nature. A token issued locally has the header claim *kid*
    starting with *A* (for *Access*) or *R* (for *Refresh*), followed by its id in the token database,
    or as a single letter in the range *[B-Z]*, less *R*. If the *kid* claim contains such an id,
    then the cryptographic key needed for validation will be obtained from the token database.
    Otherwise, the current decoding key is used.

    On success, return the token's claims (*header* and *payload*), as documented in *jwt_get_claims()*
    On failure, *errors* will contain the reason(s) for rejecting *token*.

    :param token: the token to be validated
    :param nature: prefix identifying the nature of locally issued tokens
    :param account_id: optionally, validate the token's account owner
    :param errors: incidental error messages
    :param logger: optional logger
    :return: The token's claims (*header* and *payload*) if it is valid, *None* otherwise
    """
    # initialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg="Validate JWT token")

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # extract needed data from token header
    token_header: dict[str, Any] | None = None
    try:
        token_header: dict[str, Any] = jwt.get_unverified_header(jwt=token)
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error retrieving the token's header: {exc_err}")
        errors.append(exc_err)

    if not errors:
        token_kid: str = token_header.get("kid")
        token_alg: str | None = None
        token_decoder: bytes | None = None

        # retrieve token data from database
        if nature and not (token_kid and token_kid[0:1] == nature):
            if logger:
                logger.error(f"Nature of token's 'kid' ('{token_kid}') not '{nature}'")
            errors.append("Invalid token")
        elif token_kid and len(token_kid) > 1 and \
                token_kid[0:1] in ["A", "R"] and token_kid[1:].isdigit():
            # token was likely issued locally
            where_data: dict[str, Any] = {JwtDbConfig.COL_KID: int(token_kid[1:])}
            if account_id:
                where_data[JwtDbConfig.COL_ACCOUNT] = account_id
            recs: list[tuple[str]] = db_select(sel_stmt=f"SELECT {JwtDbConfig.COL_ALGORITHM}, "
                                                        f"{JwtDbConfig.COL_DECODER} "
                                                        f"FROM {JwtDbConfig.TABLE}",
                                               where_data=where_data,
                                               engine=DbEngine(JwtDbConfig.ENGINE),
                                               errors=errors,
                                               logger=logger)
            if recs:
                token_alg = recs[0][0]
                token_decoder = b64decode(recs[0][1])
            elif errors:
                if logger:
                    logger.error(msg=f"Error retrieving the token's decoder: {'; '.join(errors)}")
            else:
                if logger:
                    logger.error(msg="Token not in the database")
                errors.append("Invalid token")
        else:
            token_alg = JwtConfig.DEFAULT_ALGORITHM.value
            token_decoder = JwtConfig.DECODING_KEY.value

        # validate the token
        if not errors:
            try:
                # raises:
                #   InvalidTokenError: token is invalid
                #   InvalidKeyError: authentication key is not in the proper format
                #   ExpiredSignatureError: token and refresh period have expired
                #   InvalidSignatureError: signature does not match the one provided as part of the token
                #   ImmatureSignatureError: 'nbf' or 'iat' claim represents a timestamp in the future
                #   InvalidAlgorithmError: the specified algorithm is not recognized
                #   InvalidIssuedAtError: 'iat' claim is non-numeric
                #   MissingRequiredClaimError: a required claim is not contained in the claimset
                payload: dict[str, Any] = jwt.decode(jwt=token,
                                                     options={
                                                         "verify_signature": True,
                                                         "verify_exp": True,
                                                         "verify_nbf": True
                                                     },
                                                     key=token_decoder,
                                                     require=["iat", "iss", "exp", "sub"],
                                                     algorithms=token_alg)
                if account_id and payload.get("sub") != account_id:
                    if logger:
                        logger.error(msg=f"Token does not belong to account '{account_id}'")
                    errors.append("Invalid token")
                else:
                    result = {
                        "header": token_header,
                        "payload": payload
                    }
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if logger:
                    logger.error(msg=f"Error decoding the token: {exc_err}")
                errors.append(exc_err)

    if not errors and logger:
        logger.debug(msg="Token is valid")

    return result


def jwt_revoke_token(account_id: str,
                     token: str,
                     errors: list[str] = None,
                     logger: Logger = None) -> bool:
    """
    Revoke the *refresh_token* associated with *account_id*.

    Revoke operations require access to a database table defined by *JWT_DB_TABLE*.

    :param account_id: the account identification
    :param token: the token to be revoked
    :param errors: incidental error messages
    :param logger: optional logger
    :return: *True* if operation could be performed, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    if logger:
        logger.debug(msg=f"Revoking token of account '{account_id}'")

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    token_claims: dict[str, Any] = jwt_validate_token(token=token,
                                                      account_id=account_id,
                                                      errors=errors,
                                                      logger=logger)
    if not errors:
        token_kid: str = token_claims["header"].get("kid")
        if token_kid[0:1] not in ["A", "R"]:
            errors.append("Invalid token")
        else:
            db_delete(delete_stmt=f"DELETE FROM {JwtDbConfig.TABLE}",
                      where_data={
                          JwtDbConfig.COL_KID: int(token_kid[1:]),
                          JwtDbConfig.COL_ACCOUNT: account_id
                      },
                      engine=DbEngine(JwtDbConfig.ENGINE),
                      errors=errors,
                      logger=logger)
    if not errors:
        result = True
    elif logger:
        logger.error(msg="; ".join(errors))

    return result


def jwt_issue_token(account_id: str,
                    nature: str,
                    duration: int,
                    lead_interval: int = None,
                    claims: dict[str, Any] = None,
                    errors: list[str] = None,
                    logger: Logger = None) -> str:
    """
    Issue or refresh, and return, a JWT token associated with *account_id*, of the specified *nature*.

    The parameter *nature* must be a single letter in the range *[B-Z]*, less *R*
    (*A* is reserved for *access* tokens, and *R* for *refresh* tokens).
    The parameter *duration* specifies the token's validity interval (at least 60 seconds).
    These claims are ignored, if specified in *claims*: *iat*, *iss*, *exp*, *jti*, *nbf*, and *sub*.

    :param account_id: the account identification
    :param nature: the token's nature, must be a single letter in the range *[B-Z]*, less *R*
    :param duration: the number of seconds for the token to remain valid (at least 60 seconds)
    :param claims: optional token's claims
    :param lead_interval: optional interval for the token to become active (in seconds)
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the JWT token data, or *None* if error
    """
    # inicialize the return variable
    result: str | None = None

    if logger:
        logger.debug(msg=f"Issuing a JWT token for '{account_id}'")

    try:
        result = __jwt_registry.issue_token(account_id=account_id,
                                            nature=nature,
                                            duration=duration,
                                            claims=claims,
                                            lead_interval=lead_interval,
                                            logger=logger)
        if logger:
            logger.debug(msg=f"Token is '{result}'")
    except Exception as e:
        # token issuing failed
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error issuing the token: {exc_err}")
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def jwt_issue_tokens(account_id: str,
                     account_claims: dict[str, Any] = None,
                     errors: list[str] = None,
                     logger: Logger = None) -> dict[str, Any]:
    """
    Issue the JWT token pair associated with *account_id*, for access and refresh operations.

    These claims are ignored, if provided in *account_claims*: *iat*, *iss*, *exp*, *jti*, *nbf*, and *sub*.
    Other claims specified therein may supercede currently registered account-related claims.

    Structure of the return data:
    {
      "access-token": <jwt-token>,
      "created-in": <timestamp>,
      "expires-in": <seconds-to-expiration>,
      "refresh-token": <jwt-token>
    }

    :param account_id: the account identification
    :param account_claims: if provided, may supercede currently registered account-related claims
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the JWT token data, or *None* if error
    """
    # inicialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg=f"Issuing a JWT token pair for '{account_id}'")

    try:
        result = __jwt_registry.issue_tokens(account_id=account_id,
                                             account_claims=account_claims,
                                             logger=logger)
        if logger:
            logger.debug(msg=f"Token data is '{result}'")
    except Exception as e:
        # token issuing failed
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error issuing the token pair: {exc_err}")
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def jwt_refresh_tokens(account_id: str,
                       refresh_token: str,
                       errors: list[str] = None,
                       logger: Logger = None) -> dict[str, Any]:
    """
    Refresh the JWT token pair associated with *account_id*, for access and refresh operations.

    The claims in *refresh-token* are used on issuing the new tokens.

    Structure of the return data:
    {
      "access-token": <jwt-token>,
      "created-in": <timestamp>,
      "expires-in": <seconds-to-expiration>,
      "refresh-token": <jwt-token>
    }

    :param errors: incidental error messages
    :param account_id: the account identification
    :param refresh_token: the base refresh token
    :param logger: optional logger
    :return: the JWT token data, or *None* if error
    """
    # inicialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg=f"Refreshing a JWT token pair for '{account_id}'")

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # assert the refresh token
    if refresh_token:
        # is the refresh token valid ?
        token_claims: dict[str, Any] = jwt_validate_token(token=refresh_token,
                                                          nature="R",
                                                          account_id=account_id,
                                                          errors=errors,
                                                          logger=logger)
        if token_claims:
            # yes, proceed
            token_kid: str = token_claims["header"].get("kid")

            # start the database transaction
            db_conn: Any = db_connect(autocommit=False,
                                      engine=DbEngine(JwtDbConfig.ENGINE),
                                      errors=errors,
                                      logger=logger)
            if db_conn:
                # delete current refresh token
                db_delete(delete_stmt=f"DELETE FROM {JwtDbConfig.TABLE}",
                          where_data={
                              JwtDbConfig.COL_KID: int(token_kid[1:]),
                              JwtDbConfig.COL_ACCOUNT: account_id
                          },
                          engine=DbEngine(JwtDbConfig.ENGINE),
                          connection=db_conn,
                          committable=False,
                          errors=errors,
                          logger=logger)

                # issue the token pair
                if not errors:
                    try:
                        result = __jwt_registry.issue_tokens(account_id=account_id,
                                                             account_claims=token_claims.get("payload"),
                                                             db_conn=db_conn,
                                                             logger=logger)
                        if logger:
                            logger.debug(msg=f"Token pair was refreshed for account '{account_id}'")
                    except Exception as e:
                        # token issuing failed
                        exc_err: str = exc_format(exc=e,
                                                  exc_info=sys.exc_info())
                        if logger:
                            logger.error(msg=f"Error refreshing the token pair: {exc_err}")
                        errors.append(exc_err)

                # wrap-up the transaction
                if errors:
                    db_rollback(connection=db_conn,
                                logger=logger)
                else:
                    db_commit(connection=db_conn,
                              errors=errors,
                              logger=logger)
                db_close(connection=db_conn,
                         logger=logger)
    else:
        # refresh token not found
        errors.append("Refresh token was not provided")

    if errors and logger:
        logger.error(msg="; ".join(errors))

    return result


def jwt_get_claims(token: str,
                   errors: list[str] = None,
                   logger: Logger = None) -> dict[str, Any] | None:
    """
    Retrieve the claims set of a JWT *token*.

    Any well-constructed JWT token may be provided in *token*, as this operation is not restricted
    to locally issued tokens. Note that neither the token's signature nor its expiration is verified.

    Structure of the returned data, for locally issued tokens:
      {
        "header": {
          "alg": "RS256",
          "typ": "JWT",
          "kid": "A1234"
        },
        "payload": {
          "valid-from": <YYYY-MM-DDThh:mm:ss+00:00>
          "valid-until": <YYYY-MM-DDThh:mm:ss+00:00>
          "birthdate": "1980-01-01",
          "email": "jdoe@mail.com",
          "exp": 1516640454,
          "iat": 1516239022,
          "iss": "my_jwt_provider.com",
          "jti": "Uhsdfgr67FGH567qwSDF33er89retert",
          "gender": "M",
          "name": "John Doe",
          "nbt": 1516249022
          "sub": "11111111111",
          "roles": [
            "administrator",
            "operator"
          ]
        }
      }

    :param token: the token to be inspected for claims
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the token's claimset, or *None* if error
    """
    # initialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg="Retrieve claims for token")

    try:
        header: dict[str, Any] = jwt.get_unverified_header(jwt=token)
        payload: dict[str, Any] = jwt.decode(jwt=token,
                                             options={"verify_signature": False})
        result = {
            "header": header,
            "payload": payload
        }
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error retrieving the token's claimsn: {exc_err}")
        if isinstance(errors, list):
            errors.append(exc_err)

    return result



import requests
import sys
from base64 import b64encode
from datetime import datetime
from logging import Logger
from pypomes_core import TZ_LOCAL, exc_format
from requests import Response
from typing import Any

# structure:
# {
#    <provider-id>: {
#      "url": <strl>,
#      "user": <str>,
#      "pwd": <str>,
#      "basic-auth": <bool>,
#      "headers-data": <dict[str, str]>,
#      "body-data": <dict[str, str],
#      "token": <str>,
#      "expiration": <timestamp>
#    }
# }
_provider_registry: dict[str, dict[str, Any]] = {}


def provider_register(provider_id: str,
                      access_url: str,
                      auth_user: str,
                      auth_pwd: str,
                      custom_auth: tuple[str, str] = None,
                      headers_data: dict[str, str] = None,
                      body_data: dict[str, str] = None) -> None:
    """
    Register an external authentication token provider.

    If specified, *custom_auth* provides key names for sending credentials (username and password, in this order)
    as key-value pairs in the body of the request. Otherwise, the external provider *provider_id* uses the standard
    HTTP Basic Authorization scheme, wherein the credentials are B64-encoded and send in the request headers.

    Optional constant key-value pairs (such as ['Content-Type', 'application/x-www-form-urlencoded']), to be
    added to the request headers, may be specified in *headers_data*. Likewise, optional constant key-value pairs
    (such as ['grant_type', 'client_credentials']), to be added to the request body, may be specified in *body_data*.

    :param provider_id: the provider's identification
    :param access_url: the url to request authentication tokens with
    :param auth_user: the basic authorization user
    :param auth_pwd: the basic authorization password
    :param custom_auth: optional key names for sending the credentials as key-value pairs in the body of the request
    :param headers_data: optional key-value pairs to be added to the request headers
    :param body_data: optional key-value pairs to be added to the request body
    """
    global _provider_registry  # noqa: PLW0602
    _provider_registry[provider_id] = {
        "url": access_url,
        "user": auth_user,
        "pwd": auth_pwd,
        "custom-auth": custom_auth,
        "headers-data": headers_data,
        "body-data": body_data,
        "token": None,
        "expiration": datetime.now(tz=TZ_LOCAL).timestamp()
    }


def provider_get_token(provider_id: str,
                       errors: list[str] | None,
                       logger: Logger = None) -> str | None:
    """
    Obtain an authentication token from the external provider *provider_id*.

    :param provider_id: the provider's identification
    :param errors: incidental error messages
    :param logger: optional logger
    """
    # initialize the return variable
    result: str | None = None

    global _provider_registry  # noqa: PLW0602
    err_msg: str | None = None
    provider: dict[str, Any] = _provider_registry.get(provider_id)
    if provider:
        now: float = datetime.now(tz=TZ_LOCAL).timestamp()
        if now > provider.get("expiration"):
            user: str = provider.get("user")
            pwd: str = provider.get("pwd")
            headers_data: dict[str, str] = provider.get("headers-data") or {}
            body_data: dict[str, str] = provider.get("body-data") or {}
            custom_auth: tuple[str, str] = provider.get("custom-auth")
            if custom_auth:
                body_data[custom_auth[0]] = user
                body_data[custom_auth[1]] = pwd
            else:
                enc_bytes: bytes = b64encode(f"{user}:{pwd}".encode())
                headers_data["Authorization"] = f"Basic {enc_bytes.decode()}"
            url: str = provider.get("url")
            try:
                # typical return on a token request:
                # {
                #   "token_type": "bearer",
                #   "access_token": <str>,
                #   "expires_in": <number-of-seconds>,
                #   optional data:
                #   "refresh_token": <str>,
                #   "refresh_expires_in": <nomber-of-seconds>
                # }
                response: Response = requests.post(url=url,
                                                   data=body_data,
                                                   headers=headers_data,
                                                   timeout=None)
                if response.status_code < 200 or response.status_code >= 300:
                    # request resulted in error, report the problem
                    err_msg = (f"POST '{url}': failed, "
                               f"status {response.status_code}, reason '{response.reason}'")
                else:
                    reply: dict[str, Any] = response.json()
                    provider["token"] = reply.get("access_token")
                    provider["expiration"] = now + int(reply.get("expires_in"))
                    if logger:
                        logger.debug(msg=f"POST '{url}': status "
                                         f"{response.status_code}, reason '{response.reason}')")
            except Exception as e:
                # the operation raised an exception
                err_msg = exc_format(exc=e,
                                     exc_info=sys.exc_info())
                err_msg = f"POST '{url}': error, '{err_msg}'"
    else:
        err_msg: str = f"Provider '{provider_id}' not registered"

    if err_msg:
        if isinstance(errors, list):
            errors.append(err_msg)
        if logger:
            logger.error(msg=err_msg)
    else:
        result = provider.get("token")

    return result


